# Author: KrorngAI Org.
# Date: December, 2025


from dataclasses import dataclass, replace
from typing import TYPE_CHECKING, List, Tuple, Union
import numpy as np

import torch
from torch import Tensor

try:
    from whisper.audio import CHUNK_LENGTH
    from whisper.decoding import (
        DecodingOptions,
        DecodingResult,
        PyTorchInference,
        MaximumLikelihoodRanker,
        GreedyDecoder,
        BeamSearchDecoder,
        SuppressBlank,
        SuppressTokens,
        ApplyTimestampRules,
        DecodingTask,
    )
except (ImportError, ModuleNotFoundError):
    print("You need to install openai-whisper package: pip install git+https://github.com/openai/whisper.git")
    raise

if TYPE_CHECKING:
    from .model import Whisper
    from .whisper import NeoWhisper

from .tokenizer import get_tokenizer, NeoTokenizer
from .nn_utils import KVCache


@torch.inference_mode()
def detect_language(
    model: "Whisper", mel: Tensor, tokenizer_name: str
) -> Tuple[Tensor, List[dict]]:
    """
    Detect the spoken language in the audio, and return them as list of strings, along with the ids
    of the most probable language tokens and the probability distribution over all language tokens.
    This is performed outside the main decode loop in order to not interfere with kv-caching.

    Returns
    -------
    language_tokens : Tensor, shape = (n_audio,)
        ids of the most probable language tokens, which appears after the startoftranscript token.
    language_probs : List[Dict[str, float]], length = n_audio
        list of dictionaries containing the probability distribution over all languages.
    """
    try:
        tokenizer = get_tokenizer(
            model.is_multilingual, num_languages=model.num_languages, encoder_name=tokenizer_name
        )
    except Exception as e:
        raise e
    if (
        tokenizer.language is None
        or tokenizer.language_token not in tokenizer.sot_sequence
    ):
        raise ValueError(
            "This model doesn't have language tokens so it can't perform lang id"
        )

    single = mel.ndim == 2
    if single:
        mel = mel.unsqueeze(0)

    # skip encoder forward pass if already-encoded audio features were given
    if mel.shape[-2:] != (model.dims.n_audio_ctx, model.dims.n_audio_state):
        mel = model.encoder(mel)

    # forward pass using a single token, startoftranscript
    n_audio = mel.shape[0]
    x = torch.tensor([[tokenizer.sot]] * n_audio).to(mel.device)  # [n_audio, 1]
    logits = model.logits(x, mel)[:, 0]

    # collect detected languages; suppress all non-language tokens
    mask = torch.ones(logits.shape[-1], dtype=torch.bool)
    mask[list(tokenizer.all_language_tokens)] = False
    logits[:, mask] = -np.inf
    language_tokens = logits.argmax(dim=-1)
    language_token_probs = logits.softmax(dim=-1).cpu()
    language_probs = [
        {
            c: language_token_probs[i, j].item()
            for j, c in zip(tokenizer.all_language_tokens, tokenizer.all_language_codes)
        }
        for i in range(n_audio)
    ]

    if single:
        language_tokens = language_tokens[0]
        language_probs = language_probs[0]

    return language_tokens, language_probs


@dataclass(frozen=True)
class NeoDecodingOptions(DecodingOptions):
    # WARN: add new option: tokenizer_name
    tokenizer_name: str = "gpt2"


class NeoPyTorchInference(PyTorchInference):
    def __init__(self, model: "NeoWhisper", initial_token_length: int):
        super().__init__(model, initial_token_length)
        self.kv_model_kwargs = {
            "num_heads": self.model.dims.n_text_head,
            "head_dim": self.model.dims.n_text_state // self.model.dims.n_text_head,
            "num_layers": self.model.dims.n_text_layer,
        }

    def logits(self, tokens: Tensor, audio_features: Tensor) -> Tensor:
        if not self.kv_cache:
            self.kv_cache, self.hooks = self.model.install_kv_cache_hooks()

        if self.kv_cache.get('neo', None) is None:
            batch_size = audio_features.shape[0]
            self.kv_cache['neo'] = KVCache(
                batch_size=batch_size,
                seq_len=self.initial_token_length,
                **self.kv_model_kwargs,
            )

        if tokens.shape[-1] > self.initial_token_length:
            # only need to use the last token except in the first forward pass
            tokens = tokens[:, -1:]

        return self.model.decoder(tokens, audio_features, kv_cache=self.kv_cache)


class NeoDecodingTask(DecodingTask):
    def __init__(self, model: "NeoWhisper", options: DecodingOptions):
        self.model = model

        tokenizer_name = options.tokenizer_name or "gpt2"
        language = options.language or "en"
        tokenizer = get_tokenizer(
            model.is_multilingual,
            num_languages=model.num_languages,
            language=language,
            task=options.task,
            encoder_name=tokenizer_name
        )
        self.tokenizer: NeoTokenizer = tokenizer
        self.options: DecodingOptions = self._verify_options(options)

        self.n_group: int = options.beam_size or options.best_of or 1
        self.n_ctx: int = model.dims.n_text_ctx
        self.sample_len: int = options.sample_len or model.dims.n_text_ctx // 2

        self.sot_sequence: Tuple[int] = tokenizer.sot_sequence
        if self.options.without_timestamps:
            self.sot_sequence = tokenizer.sot_sequence_including_notimestamps

        self.initial_tokens: Tuple[int] = self._get_initial_tokens()
        self.sample_begin: int = len(self.initial_tokens)
        self.sot_index: int = self.initial_tokens.index(tokenizer.sot)

        # inference: implements the forward pass through the decoder, including kv caching
        #TODO : this does not work for Whisper yet
        self.inference = NeoPyTorchInference(model, len(self.initial_tokens))

        # sequence ranker: implements how to rank a group of sampled sequences
        self.sequence_ranker = MaximumLikelihoodRanker(options.length_penalty)

        # decoder: implements how to select the next tokens, given the autoregressive distribution
        if options.beam_size is not None:
            self.decoder = BeamSearchDecoder(
                options.beam_size, tokenizer.eot, self.inference, options.patience
            )
        else:
            self.decoder = GreedyDecoder(options.temperature, tokenizer.eot)

        # logit filters: applies various rules to suppress or penalize certain tokens
        self.logit_filters = []
        if self.options.suppress_blank:
            self.logit_filters.append(SuppressBlank(
                self.tokenizer, self.sample_begin))
        if self.options.suppress_tokens:
            self.logit_filters.append(
                SuppressTokens(self._get_suppress_tokens()))
        if not options.without_timestamps:
            precision = CHUNK_LENGTH / model.dims.n_audio_ctx  # usually 0.02 seconds
            max_initial_timestamp_index = None
            if options.max_initial_timestamp:
                max_initial_timestamp_index = round(
                    self.options.max_initial_timestamp / precision
                )
            self.logit_filters.append(
                ApplyTimestampRules(
                    tokenizer, self.sample_begin, max_initial_timestamp_index
                )
            )


@torch.inference_mode()
def decode(
    model: "Whisper",
    mel: Tensor,
    options: DecodingOptions = DecodingOptions(),
    **kwargs,
) -> Union[DecodingResult, List[DecodingResult]]:
    """
    Performs decoding of 30-second audio segment(s), provided as Mel spectrogram(s).

    Parameters
    ----------
    model: Whisper
        the Whisper model instance

    mel: torch.Tensor, shape = (80, 3000) or (*, 80, 3000)
        A tensor containing the Mel spectrogram(s)

    options: DecodingOptions
        A dataclass that contains all necessary options for decoding 30-second segments

    Returns
    -------
    result: Union[DecodingResult, List[DecodingResult]]
        The result(s) of decoding contained in `DecodingResult` dataclass instance(s)
    """
    if single := mel.ndim == 2:
        mel = mel.unsqueeze(0)

    if kwargs:
        options = replace(options, **kwargs)

    result = NeoDecodingTask(model, options).run(mel)

    return result[0] if single else result



# Author: KrorngAI Org.
# Date: December, 2025


from typing import List, Optional

try:
    from whisper.tokenizer import (
        lru_cache,
        Tokenizer,
        tiktoken,
        LANGUAGES,
        TO_LANGUAGE_CODE
    )
except (ImportError, ModuleNotFoundError):
    print("You need to install openai-whisper package: pip install git+https://github.com/openai/whisper.git")
    raise


class NeoTokenizer(Tokenizer):
    def decode(self, token_ids: List[int], skip_special_tokens: bool = False, **kwargs) -> str:
        """
        Re-implement by adding skip_special_tokens option
        """
        token_ids = [t for t in token_ids if t < self.timestamp_begin]
        if skip_special_tokens:
            token_ids = [
                t for t in token_ids if t not in self.special_tokens.values()]
        return self.encoding.decode(token_ids, **kwargs)


@lru_cache(maxsize=None)
def get_encoding(name: str = "gpt2", num_languages: int = 99):
    """
    Inspired by https://github.com/openai/tiktoken/tree/main, Section Extending tiktoken
    And tokenizer.py of OpenAI/whisper
    """
    encoder_decoder = tiktoken.get_encoding(name)

    n_vocab = len(encoder_decoder._mergeable_ranks)
    special_tokens = {}

    specials = [
        "<|endoftext|>",
        "<|startoftranscript|>",
        *[f"<|{lang}|>" for lang in list(LANGUAGES.keys())[:num_languages]],
        "<|translate|>",
        "<|transcribe|>",
        "<|startoflm|>",
        "<|startofprev|>",
        "<|nospeech|>",
        "<|notimestamps|>",
        *[f"<|{i * 0.02:.2f}|>" for i in range(1501)],
    ]

    for token in specials:
        special_tokens[token] = n_vocab
        n_vocab += 1

    return tiktoken.Encoding(
        name=f"{name}_im",
        explicit_n_vocab=n_vocab,
        pat_str=encoder_decoder._pat_str,
        mergeable_ranks=encoder_decoder._mergeable_ranks,
        special_tokens=special_tokens,
    )


@lru_cache(maxsize=None)
def get_tokenizer(
    multilingual: bool,
    *,
    num_languages: int = 99,
    language: Optional[str] = None,
    task: Optional[str] = None,  # Literal["transcribe", "translate", None]
    encoder_name: Optional[str] = None
) -> Tokenizer:
    if language is not None:
        language = language.lower()
        if language not in LANGUAGES:
            if language in TO_LANGUAGE_CODE:
                language = TO_LANGUAGE_CODE[language]
            else:
                raise ValueError(f"Unsupported language: {language}")

    if multilingual:
        encoding_name = "multilingual"
        language = language or "en"
        task = task or "transcribe"
    else:
        encoding_name = "gpt2"
        language = None
        task = None
    if encoder_name is not None:
        encoding_name = encoder_name

    encoding = get_encoding(name=encoding_name, num_languages=num_languages)

    return NeoTokenizer(
        encoding=encoding, num_languages=num_languages, language=language, task=task
    )
